from django.db import models
from to_rest import exceptions
from to_rest import constants
from to_rest import cfg
from rest_framework.serializers import BaseSerializer
import logging
from collections import defaultdict

def restifyModel(_cls=None, *, customViewParams=None, excludeFields=None, methodFields=None):
    """
    A decorator function to include the models in the registry so that the decorated models
    are marked for restification. By restification we mean to expose REST api(s) for that 
    model.

    Parameters:
        _cls (object): The class that needs to be decorated.

        customViewParams (str): To provide class names in view_params.py for custom view parameters.

        excludeFields (list): The fields that needs to be excluded from the JSON object. provide
        fields will not be included in the serializer. If customSerializer is provided then this
        parameter will ne ignored. 

        methodFields (list): The list of methods as read only fields. This can be used to include the
        model's methods output as field. This include only those field that don't take any parameter.

    Returns:
        decorated class or function object
    """
    if customViewParams is not None and not isinstance(customViewParams, str):
        raise TypeError(constants.TYPE_ERROR_MESSAGE.format("customViewParams", "str", type(customViewParams)))
    if excludeFields is not None and not isinstance(excludeFields, list):
        raise TypeError(constants.TYPE_ERROR_MESSAGE.format("excludeFields", "list", type(excludeFields)))
    if methodFields is not None and not isinstance(methodFields, list):
        raise TypeError(constants.TYPE_ERROR_MESSAGE.format("methodFields", "list", type(methodFields)))
    def decorator_restifyModel(cls):
        """
        The decorator function that does the registry/marking.
        """
        if not issubclass(cls, models.Model):
            raise exceptions.DecoratorException(constants.NOT_A_MODEL_CLASS_MSG)
        else:
            options = None
            try:
                options = cfg.djangoToRestRegistry[cls._meta.label]
            except KeyError:
                logging.info("todjango.decorators.restifyModel.decorator_restifyModel:: Performing registration for :" + cls.__name__)
                options = defaultdict(None)
                options[constants.CUSTOM_VIEW_PARAMS] = customViewParams
                options[constants.EXCLUDE_FIELDS] = excludeFields
                options[constants.METHOD_FIELDS] = methodFields
                #options[constants.CUSTOM_SERIALIZER] = None if customViewParams is None else customViewParams.pop(constants.SERIALIZER_CLASS, None)
                cfg.djangoToRestRegistry[cls._meta.label] = options
            return cls
    
    if _cls is None:
        return decorator_restifyModel
    else:
        return decorator_restifyModel(_cls)


from to_rest import cfg
from django.apps import apps
from to_rest import constants
from to_rest import serializers as restifySerializer
from to_rest import views
from django.urls import path
from rest_framework.viewsets import ModelViewSet
from rest_framework import routers
from abc import ABC, abstractmethod
import importlib

class ViewParams(ABC):
    """
    Abstract class to provide custom view parameters. This class needs to be inherited to provide
    custom view params.
    """

    @abstractmethod
    def getParams():
        pass

def restifyApp(relativeUri):
    """
    The function to restify an app. This will iterate over the models of the appName and will
    create ModelSerializers for the models, will create viewsets and will hook them with
    routes

    Prameters:
        relativeUri: (string): The relative url for the api
    Return:
        list of urls (list)
    """

    for entity in cfg.djangoToRestRegistry:
        if cfg.djangoToRestRegistry[entity].get(constants.CUSTOM_VIEW_PARAMS,False) and isinstance(cfg.djangoToRestRegistry[entity].get(constants.CUSTOM_VIEW_PARAMS,False), str):
            temp = cfg.djangoToRestRegistry[entity][constants.CUSTOM_VIEW_PARAMS]
            appName = apps.get_model(entity)._meta.app_label
            module = importlib.import_module(appName + '.' + 'view_params')
            temp = getattr(module, temp)
            customViewParams = temp.getParams()
            cfg.djangoToRestRegistry[entity][constants.CUSTOM_SERIALIZER] = customViewParams.pop(constants.SERIALIZER_CLASS, None)
            cfg.djangoToRestRegistry[entity][constants.CUSTOM_VIEW_PARAMS] = customViewParams
    
    for entity in cfg.djangoToRestRegistry:
        model = apps.get_model(entity)
        customSerializer = cfg.djangoToRestRegistry[entity].get(constants.CUSTOM_SERIALIZER,None)
        customViewParams = cfg.djangoToRestRegistry[entity].get(constants.CUSTOM_VIEW_PARAMS,None)
        excludedFields = cfg.djangoToRestRegistry[entity].get(constants.EXCLUDE_FIELDS, None)
        methodFields = cfg.djangoToRestRegistry[entity].get(constants.METHOD_FIELDS, None)
        modelSerializer = None
        if customSerializer is None:
            modelSerializer = restifySerializer.createModelSerializers(model, excludedFields, methodFields)
            cfg.djangoToRestRegistry[entity][constants.DEFAULT_SERIALIZER] = modelSerializer
        else:
            tempSerializer = restifySerializer.createModelSerializers(model, excludedFields, methodFields)
            cfg.djangoToRestRegistry[entity][constants.DEFAULT_SERIALIZER] = tempSerializer
            modelSerializer = customSerializer
        viewSetAttributes =  views.getObjectViewSetAttributes(model, modelSerializer, customViewParams)
        cfg.djangoToRestRegistry[entity][constants.VIEW_SET_ATTRIBUTES] = viewSetAttributes
    
    router1 = routers.DefaultRouter(trailing_slash=True)
    router2 = routers.DefaultRouter(trailing_slash=False)
    for entity in cfg.djangoToRestRegistry:
        model = apps.get_model(entity)
        viewSetAttributes = cfg.djangoToRestRegistry[entity][constants.VIEW_SET_ATTRIBUTES]
        defaultActions = cfg.djangoToRestRegistry[entity].get(constants.DEFAULT_ACTIONS, None)
        if defaultActions is not None:
            for defaultAction in defaultActions:
                if defaultAction[0] == "oneToManyActionFactory":
                    serializer = None
                    customSerializer = cfg.djangoToRestRegistry[defaultAction[2]].get(constants.CUSTOM_SERIALIZER, None)
                    defaultSerializer = cfg.djangoToRestRegistry[defaultAction[2]][constants.DEFAULT_SERIALIZER]
                    if customSerializer is None:
                        serializer = defaultSerializer
                    else:
                        serializer = customSerializer
                    action = views.oneToManyActionFactory(defaultAction[1],serializer,defaultAction[3], defaultAction[4])
                    for x in action:
                        if viewSetAttributes is not None and viewSetAttributes.get(x.__name__, None) is None:
                            viewSetAttributes[x.__name__] = x
                elif defaultAction[0] == "manyToManyActionFactory":
                    action = views.manyToManyActionFactory(defaultAction[1],defaultAction[2],defaultAction[3])
                    for x in action:
                        if viewSetAttributes is not None and viewSetAttributes.get(x.__name__,None) is None:
                            viewSetAttributes[x.__name__] = x
        
        viewSet = type(constants.PROJECT_NAME_PREFIX + entity.replace('.','_') + "ViewSet", (ModelViewSet,), viewSetAttributes)
        cfg.djangoToRestRegistry[entity][constants.DEFAULT_VIEW_SET] = viewSet
        router1.register(prefix=r'{}/{}'.format(relativeUri, model._meta.label.lower().replace('.','/')), viewset=viewSet, basename=model._meta.label.lower().replace('.','_'))
        router2.register(prefix=r'{}/{}'.format(relativeUri, model._meta.label.lower().replace('.','/')), viewset=viewSet, basename=model._meta.label.lower().replace('.','_'))
        
    return router1.urls + router2.urls


from rest_framework import serializers
from to_rest import constants
from django.urls import reverse
from django.db.models.fields.related import OneToOneField, ForeignKey, ManyToManyField
from django.db.models.fields.reverse_related import OneToOneRel, ManyToOneRel, ManyToManyRel
from to_rest import cfg

def createModelSerializers(model, excludedFields, methodFields):
    """
    Method to create model serializers for the models marked for restification.

    Parameters:

        model (django.db.models.Model): The model itself

        excludedFields (list): The list of fields to be excluded

        requiredReverseRelFields (list): One to One reverse fields to be made required.  

    Returns:

        serializer (rest_framework.serializers.ModelSerializer)
    """

    def methodFieldsFactory(methodName):
        """
        Method to create seriaizer method for additional method fields

        Parameters:

            methodName (str): name of the method field

        Returns:

            func object
        """
        def func(self,object):
            temp = eval("object." + methodName + "()")
            return temp
        func.__name__ = "get_" + methodName
        return func
    
    def relationalMethodFieldsFactory(model, fieldName):
        """
        Method to create a serializer method for creating a url for the relational fields
        in case of one to many and many to many relation

        Parameters:

            model (django.db.models.Model): model for which the serializer method is created.

            fieldName (str): Name of the relational field

        Returns:

            func object
        """
        def func(self,object):
            tempName = model.__name__.lower()
            return(reverse(model._meta.label.lower().replace('.','_') + "-" + tempName + "-" + fieldName + "-list", args = [object.pk] ))
        func.__name__ = "get_" + fieldName
        return func

    
    fields = []
    relationalFields = []
    for field in model._meta.get_fields():
        if excludedFields is not None and field.name in excludedFields:
            continue
        elif field.is_relation:
            if cfg.djangoToRestRegistry.get(field.related_model._meta.label, False):
                relationalFields.append(field)
        else:
            fields.append(field.name)
    if methodFields is not None:
        fields.extend(methodFields)
    #create meta attributes for the meta class
    metaAttributes = {}
    metaAttributes["model"] = model
    
    #create attributes for serializer class
    serializerAttribute = {}
    defaultActions = []
    oneToOneRelations = []
    for relationalField in relationalFields:
        if isinstance(relationalField, OneToOneField):
            fields.append(relationalField.name)
        elif isinstance(relationalField, OneToOneRel):
            temp = relationalField.related_name if relationalField.related_name is not None else relationalField.name
            fields.append(temp)
            oneToOneRelations.append(temp)
        elif isinstance(relationalField, ForeignKey):
            fields.append(relationalField.name)
        elif isinstance(relationalField, ManyToOneRel):
            temp = relationalField.related_name if relationalField.related_name is not None else relationalField.name + "_set"
            serializerAttribute[temp] = serializers.SerializerMethodField(read_only=True)
            serializerAttribute["get_"+temp] = relationalMethodFieldsFactory(model, temp)
            fields.append(temp)
            defaultActions.append(("oneToManyActionFactory", model, relationalField.related_model._meta.label, relationalField, temp))
        elif isinstance(relationalField, ManyToManyRel):
            temp = relationalField.related_name if relationalField.related_name is not None else relationalField.name + "_set"
            serializerAttribute[temp] = serializers.SerializerMethodField(read_only=True)
            serializerAttribute["get_"+temp] = relationalMethodFieldsFactory(model, temp)
            fields.append(temp)
            defaultActions.append(("manyToManyActionFactory", model, relationalField,temp)) 
        elif isinstance(relationalField, ManyToManyField):
            temp = relationalField.name
            serializerAttribute[temp] = serializers.SerializerMethodField(read_only=True)
            serializerAttribute["get_"+temp] = relationalMethodFieldsFactory(model, temp)
            fields.append(temp)
            defaultActions.append(("manyToManyActionFactory", model, relationalField,temp))
            #in the above if conditions, certain tuples are added to defaultActions to create
            #corresponding action for the one to many and many to many fields
    cfg.djangoToRestRegistry[model._meta.label][constants.DEFAULT_ACTIONS] = defaultActions
    metaAttributes["fields"] = fields
    meta = type("Meta", (object,), metaAttributes)
    serializerAttribute["Meta"] = meta #create meta class for the serializer class
    for each in oneToOneRelations:
        serializerAttribute[each] = serializers.PrimaryKeyRelatedField(many=False, read_only=True)

    if methodFields is not None:
        for methodField in methodFields:
            serializerAttribute[methodField] = serializers.SerializerMethodField(read_only=True)
            serializerAttribute["get_" + methodField] = methodFieldsFactory(methodField)
    
    modelSerializer = type(constants.PROJECT_NAME_PREFIX + model._meta.label.replace('.','_'), (serializers.ModelSerializer,), serializerAttribute)
    
    return modelSerializer


from collections import defaultdict
djangoToRestRegistry = defaultdict(None) #a dictionary which is used as a registry to store 
#all the components of the project like serializers, viewsets, actions, custom parameters, etc


import traceback
from django.conf import settings
from rest_framework.response import Response
from rest_framework import status
from rest_framework.decorators import action
from rest_framework import serializers
from django.db.models.fields.reverse_related import ManyToManyRel
from django.db.models.fields.related import ForeignKey
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.filters import SearchFilter, OrderingFilter
from to_rest import constants
from rest_framework.viewsets import ModelViewSet
from to_rest import cfg
from django.shortcuts import get_object_or_404

REST_FRAMEWORK_SETTINGS = None
try:
    REST_FRAMEWORK_SETTINGS = settings.REST_FRAMEWORK
except Exception:
    pass

def listMethodFactory(modelName):
    def list(self, request, *args, **kwargs):
        objects = self.filter_queryset(self.get_queryset())
        if len(objects) == 0:
            headers = {}
            headers[constants.CONTENT_MESSAGE] = constants.NO_OBJECT_EXISTS.format("related "+modelName)
            return Response(status=status.HTTP_204_NO_CONTENT, headers=headers)
        else:
            page = self.paginate_queryset(objects)
            if page is not None:
                serializer =  self.get_serializer(page, many=True)
                return self.get_paginated_response(serializer.data)
            serializer = self.get_serializer(objects, many=True)
            return Response(serializer.data)
    return list

def isDefaultSerializer(serializer):
    """
    Function to check if a serializer is default or custom.

    Parameters:
        serializer (BaseSerializer): the serializer object

    Returns:
        boolean
    """
    return serializer.__name__.startswith(constants.PROJECT_NAME_PREFIX)


def getTempViewSet(queryset, childModel, childSerializer, viewParams, extras):
    """
    Function to create temporary views for filtering and permission purposes for actions as
    actions do not accept additional filterset_class or filterset_fields.

    Parameters:
        queryset: the quesryset for the view
        childModel (django.db.models.Model) : the model object
        childSerializer (rest_framework.serializers.BaseSerializer) : the serializer
        viewParams (dict) : a dictionary for the view attributes
    
    Returns:
        ViewSet
    """
    
    defaultFilterBackends = [DjangoFilterBackend, SearchFilter, OrderingFilter]
    #set defaults
    attributes = dict()
    attributes[constants.QUERYSET] = queryset
    attributes[constants.SERIALIZER_CLASS] = childSerializer
    #attributes[constants.FILTER_BACKENDS] = defaultFilterBackends if REST_FRAMEWORK_SETTINGS is None else REST_FRAMEWORK_SETTINGS.get(constants.DEFAULT_FILTER_BACKENDS, defaultFilterBackends)
    if REST_FRAMEWORK_SETTINGS is None or REST_FRAMEWORK_SETTINGS.get(constants.DEFAULT_FILTER_BACKENDS, None) is None:
        attributes[constants.FILTER_BACKENDS] = defaultFilterBackends
    if isDefaultSerializer(childSerializer):
        attributes[constants.FILTERSET_FIELDS] = [x.name for x in childModel._meta.get_fields() if (not x.is_relation or x.many_to_one)]
        attributes[constants.SEARCH_FIELDS] = [x.name for x in childModel._meta.get_fields() if (not x.is_relation)] #no relational field by default as there could be large number of relational lookups possible
        attributes[constants.ORDERING_FIELDS] = [x.name for x in childModel._meta.get_fields() if (not x.is_relation or x.many_to_one)]
    #update with custom params
    if viewParams is not None:
        attributes.update(viewParams)
    if attributes.get(constants.FILTERSET_CLASS, False):
        if attributes.get(constants.FILTERSET_FIELDS, False):
            del attributes[constants.FILTERSET_FIELDS]#remove filterset_filds if filterset_class exists
    
    if extras is not None and extras.get('create', False):
        if not attributes.get('create', False):
            attributes['create'] = extras['create']
    
    if extras is not None and extras.get('list', False):
        if not attributes.get('list', False):
            attributes['list'] = extras['list']
    
    if extras is not None and extras.get('update', False):
        if not attributes.get('update', False):
            attributes['update'] = extras['update']
    
    if extras is not None and extras.get('get_object', False):
        if not attributes.get('get_object', False):
            attributes['get_object'] = extras['get_object']

    return type(constants.PROJECT_NAME_PREFIX + "temp_" + childModel._meta.label.replace('.','_'), (ModelViewSet,), attributes)


def oneToManyActionFactory(parentModel,childSerializer, field, relatedName):
    """
    Method to create actions for view set for one to many relationship. Creation and updation
    of relationship can be handled from the other side of the relationship. Hence, no methods 
    for put,patch,delete. Since, there are use cases where an element of an entity can be
    related to an element of another entity more than once with other additional info. Hence,
    having a retreive method makes no sense by default. However, if required, custom actions
    can be provided as seen in decorators.py.

    Parameters:

        parentModel (django.db.models.Model): The model in the one to many side.

        childSerializer (rest_framework.serializers.ModelSerializer): Serializer for model in
        the other side.

        field (django.db.models.fields.Field): The field involved in the relationship.

        relatedName (str): The related name for the field.

    Returns:

        tuple of methods (tuple)

    """
    childModel = field.related_model
    parentModelName = parentModel.__name__
    childCustomViewParams = None
    if cfg.djangoToRestRegistry.get(childModel._meta.label, False):
        if cfg.djangoToRestRegistry[childModel._meta.label].get(constants.CUSTOM_VIEW_PARAMS, False):
            childCustomViewParams = cfg.djangoToRestRegistry[childModel._meta.label][constants.CUSTOM_VIEW_PARAMS]    


    def funcRelatedList(self,request,pk=None, *args, **kwargs):
        parentObject = get_object_or_404(parentModel, pk=pk)
        childObjects = eval("parentObject.{}.all()".format(relatedName))
        tempViewSet = getTempViewSet(childObjects, childModel, childSerializer, childCustomViewParams, {'list': listMethodFactory(childModel.__name__)})
        tempView = tempViewSet.as_view({'get': 'list'})
        return tempView(request._request,*args,**kwargs)

    funcRelatedList.__name__ = constants.ONE_TO_MANY_LIST_ACTION + relatedName
    funcRelatedList = action(detail=True, methods=['get'], url_path=relatedName, url_name=parentModelName.lower() + "-" + relatedName +"-list")(funcRelatedList)
    return (funcRelatedList,)

def manyToManyActionFactory(parentModel, field, relatedName):
    """
    Method to create actions for view set for many to many relationship. Since, there are 
    use cases where an element of an entity can be related to an element of another entity 
    more than once with other additional info. Hence, having a retreive method makes no 
    sense by default. Also, for the same reason, for update, partial_update and delete
    the primary key of the through model will be used as the path parameter for the nested url.
    However, if required, custom actions can be provided as seen in decorators.py.

    Parameters:

        parentModel (django.db.models.Model): The model in the one to many side.

        field (django.db.models.fields.Field): The field involved in the relationship.

        relatedName (str): The related name for the field.
    
    Returns:

        tuple of methods (tuple)

    """

    throughModel = eval("parentModel.{}.through".format(relatedName))
    throughModelName = throughModel.__name__
    parentModelName = parentModel.__name__
    metaAttributes = dict()
    metaAttributes["model"] = throughModel
    metaAttributes["fields"] = [x.name for x in throughModel._meta.get_fields()]
    meta = type("Meta", (object,), metaAttributes)
    serializerAttribute = {"Meta": meta}
    throughSerializer = type(constants.PROJECT_NAME_PREFIX + throughModelName+"Serializer", (serializers.ModelSerializer,), serializerAttribute)

    if cfg.djangoToRestRegistry.get(throughModel._meta.label, False):
        if cfg.djangoToRestRegistry[throughModel._meta.label].get(constants.CUSTOM_SERIALIZER, cfg.djangoToRestRegistry[throughModel._meta.label].get(constants.DEFAULT_SERIALIZER, False)):       
            throughSerializer = cfg.djangoToRestRegistry[throughModel._meta.label].get(constants.CUSTOM_SERIALIZER, cfg.djangoToRestRegistry[throughModel._meta.label][constants.DEFAULT_SERIALIZER])
    
    viewParams = None

    if cfg.djangoToRestRegistry.get(throughModel.__name__, False):
        if cfg.djangoToRestRegistry[throughModel.__name__].get(constants.CUSTOM_VIEW_PARAMS, False):
            viewParams = cfg.djangoToRestRegistry[throughModel.__name__][constants.CUSTOM_VIEW_PARAMS]

    def funcRelatedList(self,request,pk=None, *args,**kwargs):
        #Kind of list view
        if self.request.method == "GET":
            #The listing will be of the through objects as it present more information about the relation when there
            # are additional information.
            parentObject = get_object_or_404(parentModel, pk=pk) #just to raise 404
            filter_param = field.field.m2m_reverse_field_name() + "_" + field.field.m2m_reverse_target_field_name() if isinstance(field,ManyToManyRel) else field.m2m_field_name() + "_" + field.m2m_target_field_name()
            throughObjects = eval("throughModel.objects.filter({}=pk)".format(filter_param))
            tempViewSet = getTempViewSet(throughObjects, throughModel, throughSerializer, viewParams, {'list': listMethodFactory(throughModelName)})
            tempView = tempViewSet.as_view({'get': 'list'})
            return tempView(request._request,*args,**kwargs)
            
        elif self.request.method == "POST":
            #To create new relation ship, through objects would be used as it contains prmary key
            #of both sides of the relationship
            parentObject = get_object_or_404(parentModel, pk=pk) #just to raise 404
            providedData = request.data.copy()
            parentObjectField = field.field.m2m_reverse_field_name()  if isinstance(field,ManyToManyRel) else field.m2m_field_name()
            providedData[parentObjectField] = pk
            request._full_data = providedData
            throughObjects = parentModel.objects.all()
            def create(self, request, *args, **kwargs):
                serializer = self.get_serializer(data=providedData)
                serializer.is_valid(raise_exception=True)
                self.perform_create(serializer)
                headers = self.get_success_headers(serializer.data)
                return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)
            tempViewSet = getTempViewSet(throughObjects, throughModel, throughSerializer, viewParams, {'create': create})
            tempView = tempViewSet.as_view({'post': 'create'})
            return tempView(request._request, *args, **kwargs)
        else:
            return Response(status=status.HTTP_400_BAD_REQUEST)

    
    funcRelatedList.__name__ = constants.MANY_TO_MANY_LIST_ACTION + relatedName
    funcRelatedList = action(detail=True, methods=['get', 'post'], url_path=relatedName, url_name=parentModelName.lower() + "-" + relatedName +"-list")(funcRelatedList)

    def funcRelatedDetail(self,request,childPk,pk=None,*args,**kwargs):
        #kind of detail view
        if self.request.method == 'GET':
            parentObject = get_object_or_404(parentModel, pk=pk) #just to raise 404
            throughObject = get_object_or_404(throughModel, pk=childPk)
            def get_object(self):
                return throughObject
            throughObjects = throughModel.objects.all()
            tempViewSet = getTempViewSet(throughObjects, throughModel, throughSerializer, viewParams, {'get_object': get_object})
            tempView = tempViewSet.as_view({'get': 'retrieve'})
            return tempView(request._request, *args, **kwargs)
        elif self.request.method == "PUT":
            #For, updating a relationship, through object will be used. For, that reason, the primary
            #key of the through object will be used as path parameter in the nested url. To get the primary
            #key of the through object filter can be used in with the url of list view (GET)
            parentObject = get_object_or_404(parentModel, pk=pk) #just to raise 404
            parentObjectField = field.field.m2m_reverse_field_name()  if isinstance(field,ManyToManyRel) else field.m2m_field_name()
            throughObject = get_object_or_404(throughModel, pk=childPk)
            providedData = request.data.copy()
            providedData[parentObjectField] = pk
            def get_object(self):
                return throughObject
            def update(self, request, *args, **kwargs):
                partial = kwargs.pop('partial', False)
                instance = self.get_object()
                serializer = self.get_serializer(instance, data=providedData, partial=partial)
                serializer.is_valid(raise_exception=True)
                self.perform_update(serializer)

                if getattr(instance, '_prefetched_objects_cache', None):
                    # If 'prefetch_related' has been applied to a queryset, we need to
                    # forcibly invalidate the prefetch cache on the instance.
                    instance._prefetched_objects_cache = {}

                return Response(serializer.data)
            throughObjects = throughModel.objects.all()
            tempViewSet = getTempViewSet(throughObjects, throughModel, throughSerializer, viewParams, {'update': update, 'get_object': get_object})
            tempView = tempViewSet.as_view({'put': 'update'})
            return tempView(request._request, *args, **kwargs)
        elif self.request.method == "PATCH":
            #For, updating a relationship, through object will be used. For, that reason, the primary
            #key of the through object will be used as path parameter in the nested url. To get the primary
            #key of the through object filter can be used in with the url of list view (GET)
            parentObject = get_object_or_404(parentModel, pk=pk) #just to raise 404
            parentObjectField = field.field.m2m_reverse_field_name()  if isinstance(field,ManyToManyRel) else field.m2m_field_name()
            throughObject = get_object_or_404(throughModel, pk=childPk)
            providedData = request.data.copy()
            providedData[parentObjectField] = pk
            def get_object(self):
                return throughObject
            def update(self, request, *args, **kwargs):
                partial = kwargs.pop('partial', False)
                instance = self.get_object()
                serializer = self.get_serializer(instance, data=providedData, partial=partial)
                serializer.is_valid(raise_exception=True)
                self.perform_update(serializer)

                if getattr(instance, '_prefetched_objects_cache', None):
                    # If 'prefetch_related' has been applied to a queryset, we need to
                    # forcibly invalidate the prefetch cache on the instance.
                    instance._prefetched_objects_cache = {}

                return Response(serializer.data)
            throughObjects = throughModel.objects.all()
            tempViewSet = getTempViewSet(throughObjects, throughModel, throughSerializer, viewParams, {'update': update, 'get_object': get_object})
            tempView = tempViewSet.as_view({'patch': 'partial_update'})
            return tempView(request._request, *args, **kwargs)
        elif self.request.method == "DELETE":
            #For, deleting a relationship, through object will be used. For, that reason, the primary
            #key of the through object will be used as path parameter in the nested url. To get the primary
            #key of the through object filter can be used in with the url of list view (GET)
            parentObject = get_object_or_404(parentModel, pk=pk) #just to raise 404
            throughObject = get_object_or_404(throughModel, pk=childPk)
            def get_object(self):
                return throughObject
            throughObjects = throughModel.objects.all()
            tempViewSet = getTempViewSet(throughObjects, throughModel, throughSerializer, viewParams, {'get_object': get_object})
            tempView = tempViewSet.as_view({'delete': 'destroy'})
            return tempView(request._request, *args, **kwargs)
            # self.perform_destroy(throughObject)
            # return Response(status=status.HTTP_204_NO_CONTENT)
        else:
            return Response(status=status.HTTP_400_BAD_REQUEST)
    
    funcRelatedDetail.__name__ = constants.MANY_TO_MANY_DETAIL_ACTION + relatedName
    funcRelatedDetail = action(detail=True, methods=['get','put','patch','delete'], url_path=relatedName + "/(?P<childPk>.+)", url_name=parentModelName.lower() + "-" + relatedName +"-detail")(funcRelatedDetail)
    
    return (funcRelatedList, funcRelatedDetail)
    
def getObjectViewSetAttributes(model, modelSerializer, customViewParams):
    """
    Method to create all the attributes for view set.

    Parameters:

        model (django.db.models.Model) : The model for which the Viewset needs to be created

        modelSerializer (rest_framework.serialzers.ModelSerializer) : Corresponding model serializer fo the model.

        customViewParams (dict) : Dictionary of custom view attributes.
    
    Returns:

        attributes (dict): dictionary of view attributes
    """
    viewClassName = model.__name__ + "ViewSet"
    def list(self, request, *args, **kwargs):
        objects = self.filter_queryset(self.get_queryset())
        if len(objects) == 0:
            headers = {}
            headers[constants.CONTENT_MESSAGE] = constants.NO_OBJECT_EXISTS.format(model.__name__)
            return Response(status=status.HTTP_204_NO_CONTENT, headers=headers)
        else:
            page = self.paginate_queryset(objects)
            if page is not None:
                serializer =  self.get_serializer(page, many=True)
                return self.get_paginated_response(serializer.data)
            serializer = self.get_serializer(objects, many=True)
            return Response(serializer.data)
    
    attributes = {}
    #add defaults
    defaultFilterBackends = [DjangoFilterBackend, SearchFilter, OrderingFilter]
    attributes["queryset"] = model.objects.all()
    attributes["serializer_class"] = modelSerializer
    attributes["list"] = list
    #attributes[constants.FILTER_BACKENDS] = defaultFilterBackends if REST_FRAMEWORK_SETTINGS is None else REST_FRAMEWORK_SETTINGS.get(constants.DEFAULT_FILTER_BACKENDS, defaultFilterBackends)
    if REST_FRAMEWORK_SETTINGS is None or REST_FRAMEWORK_SETTINGS.get(constants.DEFAULT_FILTER_BACKENDS, None) is None:
        attributes[constants.FILTER_BACKENDS] = defaultFilterBackends
    if isDefaultSerializer(modelSerializer):
        attributes[constants.FILTERSET_FIELDS] = [x.name for x in model._meta.get_fields() if (not x.is_relation or x.many_to_one)]
        attributes[constants.SEARCH_FIELDS] = [x.name for x in model._meta.get_fields() if (not x.is_relation)] #no relational field by default as there could be large number of relational lookups possible
        attributes[constants.ORDERING_FIELDS] = [x.name for x in model._meta.get_fields() if (not x.is_relation or x.many_to_one)]
    #update with custom params
    if customViewParams is not None:
        attributes.update(customViewParams)
    if attributes.get(constants.FILTERSET_CLASS, False):
        if attributes.get(constants.FILTERSET_FIELDS, False):
            del attributes[constants.FILTERSET_FIELDS]#remove filterset_filds if filterset_class exists

    return attributes


#options
LIST_METHOD = "LIST_METHOD"
RETREIVE_METHOD = "RETREIVE_METHOD"
CREATE_METHOD = "CREATE_METHOD"
UPDATE_METHOD = "UPDATE_METHOD"
PARTIAL_UPDATE_METHOD = "PARTIAL_UPDATE_METHOD"
DESTROY_METHOD = "DESTROY_METHOD"
GET_OBJECT_METHOD = "GET_OBJECT_METHOD"
GET_QUERYSET_METHOD = "GET_QUERYSET_METHOD"
CUSTOM_VIEW_PARAMS = "CUSTOM_VIEW_PARAMS"
EXCLUDE_FIELDS = "EXCLUDE_FIELDS"
FILTER_SETTINGS = "FILTER_SETTINGS"
METHOD_FIELDS = "METHOD_FIELDS"
CUSTOM_SERIALIZER = "CUSTOM_SERIALIZER"
DEFAULT_SERIALIZER = "DEFAULT_SERIALIZER"
VIEW_SET_ATTRIBUTES = "VIEW_SET_ATTRIBUTES"
REQUIRED_REVERSE_REL_FIELDS = "REQUIRED_REVERSE_REL_FIELDS"
DEFAULT_ACTIONS = "DEFAULT_ACTIONS"
CUSTOM_ACTIONS = "CUSTOM_ACTIONS"
DEFAULT_VIEW_SET = "DEFAULT_VIEW_SET"

#settings constants
DEFAULT_FILTER_BACKENDS = "DEFAULT_FILTER_BACKENDS"

#other view attributes
QUERYSET = "queryset"
SERIALIZER_CLASS = "serializer_class"
FILTERSET_CLASS = "filterset_class"
FILTERSET_FIELDS = "filterset_fields"
SEARCH_FIELDS = "search_fields"
ORDERING_FIELDS = "ordering_fields"
FILTER_BACKENDS = "filter_backends"
ORDERING = "ordering"
AUTHENTICATION_CLASSES = "authentication_classes"
PERMISSION_CLASSES = "permission_classes"
THROTTLE_SCOPE = "throttle_scope"
#affixes
ONE_TO_MANY_LIST_ACTION = "oneToManyList_"
MANY_TO_MANY_LIST_ACTION = "manyToManyList_"
MANY_TO_MANY_DETAIL_ACTION = "manyToManyDetail_"
PROJECT_NAME_PREFIX = "DjangoToRest_"

#header keys
CONTENT_MESSAGE = "Content-Message"


#messages
NOT_A_MODEL_CLASS_MSG = "The decorated class must be a sub class django.db.models.Model"
NO_OBJECT_EXISTS = "No {} object exists"
TYPE_ERROR_MESSAGE = "{}: Expected {} but got {}"